/* bridge.c
 *
 *
 *    this file is part of:
 *
 *    Cyclone   ---   Loop Giant
 *
 */
 
#include <stdio.h>
#include <stdlib.h>
#include <stddef.h>
#include <pthread.h>
#include <string.h>
#include <unistd.h>
#include <stdbool.h>
#include <math.h>

#include "bridge.h"

int bridge(const char *path, struct loopdata *loop) {
    
  loop->sound_file = sf_open(path, SFM_READ, &loop->sfnfo);
  if(!loop->sound_file) {
    fprintf(stderr, "BRIDGE: Error reading file: %s\n", path);
    return 1;
  }
  
  strcpy(loop->path, path);
  loop->channels = loop->sfnfo.channels;
  if(loop->channels < 1) {
    fprintf(stderr, "BRIDGE: Illegal number of channels in file: %d\n",
	    loop->channels);
    return 1;
  }
  
  return 0;
}

const char *errorstring(struct loopdata *loop) {
  const char* errorstr = sf_strerror(loop->sound_file);
  sf_close(loop->sound_file);
  return errorstr;
}

int bridgesetindex(struct loopdata *loop, int li, int ls) {
  /* Set index */
  loop->index = li;
  loop->session = ls;
}
  
int bridgeopenbuffer(struct loopdata *loop, int li, int ls) {
  /* Set index */
  loop->index = li;
  loop->session = ls;
  loop->start = 0;
  loop->end = 0;
  
  /* Allocate buffers. */
  loop->buffer_samples = buffer_frames * loop->channels;
  loop->buffer_bytes = loop->buffer_samples * sizeof(float);
  loop->d_buffer = malloc(loop->buffer_bytes);
  loop->e_buffer = malloc(loop->buffer_bytes); 
  loop->j_buffer = malloc(loop->buffer_bytes); 
  loop->k_buffer = malloc(loop->buffer_bytes);
  loop->rb = jack_ringbuffer_create(loop->buffer_bytes);
  memset(loop->e_buffer, 0, loop->buffer_bytes);
  /* Setup sample rate conversion. */
  int aio;
  loop->src = src_callback_new(read_input_rb, SRC_SINC_FASTEST, loop->channels, &aio, loop);
  if(!loop->src) {
    fprintf(stderr,"BRIDGE: Sample rate conversion setup failed: %s\n",
	    src_strerror(aio));
    return 1;
  }
  
  /* Inform the user of sample-rate mismatch. */
  int osr = jack_get_sample_rate(jack_client);
  int isr = loop->sfnfo.samplerate;
  if(osr != isr) {
    if (debug)
      fprintf(stdout, "BRIDGE: resampling, sample rate of file != server, %d != %d\n", isr, osr);
  }
  
  /* Create communication pipe. */
  aio = pipe(loop->pipe);
  if(aio) {
    fprintf(stderr, "BRIDGE:Failed while creating pipes.\n");
    return 1;
  }
    
  /* Start disk thread, the priority number is a random guess.... */
  jack_client_create_thread (jack_client, &(loop->disk_thread), 50, true, disk_proc, loop);
    
  if (debug)
    fprintf(stdout,"STARTING: %s\t-\tchannels: %d\n", loop->path, loop->channels);
  
  if (loop->position > 0)
    resetbridge(loop);
  
  return 0;
}

void setSamplerate(long sr) {
  sample_rate = sr;
}

void setJackClient(jack_client_t *cli) {
  jack_client = cli;
}

void setMidiSeq(snd_seq_t *seqcy, int port) {
  seq = seqcy;
  midi_out = port;
}

long read_input_rb(void *PTR, float **buf) {
  struct loopdata *loop = PTR;
  int nsamples = loop->channels * 64;
  int nbytes = (size_t)nsamples * sizeof(float);

  int err = jack_ringbuffer_read(loop->rb, (char *)loop->k_buffer, nbytes);
  err /= loop->channels * sizeof(float);
  *buf = loop->k_buffer;
  /* SRC locks up if we return zero here, return a silent frame */
  if (err == 0) {
    if (debug) {
      fprintf(stdout, "BRIDGE: ringbuffer empty, silencing %s\n", loop->path);
    }
    memset(loop->k_buffer, 0, (size_t)nsamples * sizeof(float));
    err = 64;
  }

  return (long)err;
}

void *disk_proc(void *PTR) {
  struct loopdata *loop = (struct loopdata *)PTR;
  while (true) {
    if (loop->position >= loop->sfnfo.frames) {
      sf_count_t zzz = sf_seek(loop->sound_file, (sf_count_t)0, SEEK_SET);
      if(zzz == -1) {
	fprintf(stderr, "BRIDGE: seek request failed\n");
      }
      quicktrigger = 0;
      loop->position = 0;
      loop->end = 1;
      endloop();
      loop->end = 0;
    }
    /* Wait for write space at the ring buffer. */
    int nbytes = 32 * sizeof(float) * loop->channels;
    nbytes = jack_ringbuffer_wait_for_write(nbytes, loop);
    if (loop->start)
      loop->position += nbytes / (sizeof(float) * loop->channels);
    
    if (!loop->async) {
      currentposition = loop->position;
    } else {
      currentasyncposition = loop->position;
    }
    
    /* Do not overflow the local buffer. */
    if(nbytes > loop->buffer_bytes) {
      fprintf(stderr, "BRIDGE: impossible condition, write space.\n");
      nbytes = loop->buffer_bytes;
      exit(1);
    }

    /* Read sound file data, which *must* be frame aligned. */
    int nframes =(nbytes / sizeof(float))/ loop->channels;
    int nsamples = nframes * loop->channels;
    
    
    /* Write data to ring buffer. */
    if (loop->async)
      quickavolume = quick_async(loop);
    
    if (loop->start) {
      sf_count_t errsf = sf_read_float(loop->sound_file, loop->d_buffer, (sf_count_t)nsamples);
      
      if(errsf < 0) {
	fprintf(stderr, "BRIDGE: Libsndfile failed.\n");
	exit(1);    
      }
      
      if(errsf == 0) {
	return NULL;
      }
      jack_ringbuffer_write(loop->rb, (char *)loop->d_buffer, (size_t)errsf * sizeof(float));
    } else {
      jack_ringbuffer_write(loop->rb, (char *)loop->e_buffer, (size_t)32 * sizeof(float));
    }
    
    
  }
  fprintf(stderr, "BRIDGE: exit from disk_proc() :-( :%s\n", loop->path);  
  return NULL;
}

int jack_ringbuffer_wait_for_write(int nbytes, struct loopdata *loop)
{ 
  int fd = loop->pipe[0];
  int space = (int)jack_ringbuffer_write_space(loop->rb);
  while(space < nbytes) {
    char b;
    if(read(fd, &b, 1)== -1) {
      fprintf (stderr, "BRIDGE: %s: error reading communication pipe\n", __func__);      
      exit(1);
    }
    space = (int) jack_ringbuffer_write_space(loop->rb);
  }
  return space;
}

void deletebridge(struct loopdata *loop) {
  //resetbridge(loop);
  if (debug)
    printf("STOPPING: %s\n",loop->path);
  pthread_cancel(loop->disk_thread);
  jack_ringbuffer_free(loop->rb);
  close(loop->pipe[0]);
  close(loop->pipe[1]);
  free(loop->d_buffer);
  free(loop->j_buffer);
  free(loop->k_buffer);
  src_delete(loop->src);
}

void resetbridge(struct loopdata *loop) {
  jack_ringbuffer_t *rinbuf = jack_ringbuffer_create(loop->buffer_bytes);
  long space = jack_ringbuffer_read_space(loop->rb);
  jack_ringbuffer_write(rinbuf, (char *)loop->e_buffer,(size_t) space);
  jack_ringbuffer_free(loop->rb);
  loop->rb = rinbuf;
  loop->position = 0;
  loop->start = 0;
  sf_count_t zzz = sf_seek(loop->sound_file, (sf_count_t)0, SEEK_SET);
  if(zzz == -1) {
    fprintf(stderr, "BRIDGE: seek request failed\n");
  }
}

void firststartbridge(struct loopdata *loop) {
  if (debug)
    fprintf(stdout," ---- FirstStart ---- \n");
  loop->start = 1;
}

void closefile(struct loopdata *loop) {
  if (debug)
    printf("CLOSING: %s\n",loop->path);  
  sf_close(loop->sound_file);  
}

void modifyvolume(struct loopdata *loop, int vol) {
  int old = 0;
  loop->volume = vol;
  if ((currentindex == loop->index) && (currentsession == loop->session)) {
    if ((nexti == 128) || (!nextlooptrig)) {
      quickvolume = vol;
      currentLoop.volume = vol;
    } else {
      old = 1;
    }
  } else {
    old = 1;
  }
  if (old && nextlooptrig) {
    if ((oldindex == loop->index) && (oldsession == loop->session)) {
      quickvolume = vol;
      currentLoop.volume = vol;      
    }
  }
  
  if ((loops[asyncLoopindex][asyncLoopsession].index == loop->index) && (loops[asyncLoopindex][asyncLoopsession].session == loop->session)) {
    quickavolume = vol;
    asyncLoop.volume = vol;
  }
}

void setCurrentLoop(struct loopdata loop) {
  quicktrigger = 0;
  process_sl_trigger();
  currentsamples = loop.sfnfo.frames;
  currentLoop = loop;
}

void unsetCurrentLoop() {
  quickvolume = 0;
  resetbridge(&loops[currentloopindex][currentloopsession]);
}

void enableLooping() {
  currentEnabled = 1;
}

void disableLooping() {
  if (oldstate > -1)
    resetbridge(&loops[currentloopindex][currentloopsession]);
  if (nextlooptrig == 1)
    resetbridge(&loops[oldindex][oldsession]);
  nextlooptrig = 0;
  currentEnabled = 0;
}

void totallyDisableLooping() {
  disableLooping();
  if (asyncEnabled) {
    resetbridge(&loops[asyncLoopindex][asyncLoopsession]);
    currentasyncposition = 0;
    quickavolume = 0;
  }
  asyncEnabled = 0;
}

void setDebugBridge(int dbg) {
  debug = dbg;
}

float quick_async(struct loopdata *loop) {
  long nframes = (quickmsec * loop->sfnfo.samplerate) / 1000;
  float vol = (float)loop->volume;
  if (nframes >= currentasyncposition) {
    vol = (((float)currentasyncposition / (float)nframes) * (float)loop->volume);
  }
  if (currentasyncposition >= (loop->sfnfo.frames - nframes)) {
    vol = (((float)abs(currentasyncposition - (float)loop->sfnfo.frames) / nframes) * (float)loop->volume);
  }
  return vol;
}

float quick_volume(struct loopdata *loop) {
  if (quicktrigger == 0) {
    quicktrigger = 1;
    quickvolume = loop->volume;
    return 0;
  } 
  long nframes = (quickmsec * loop->sfnfo.samplerate) / 1000;
  float vol = (float)quickvolume;
  if (quicktrigger == 1) {
    if (nframes >= currentposition) {
      vol = (((float)currentposition / (float)nframes) * (float)quickvolume);
    } else {
      quicktrigger = 2;
    }
  }
  if (quicktrigger == 2) {
    if (currentposition >= (currentsamples - nframes)) {
      vol = (((float)abs(currentposition - currentsamples) / (float)nframes) * (float)quickvolume);
    }
  }
  return vol;
}

void endloop() {
  if (endasync())
    return;
  while (changeend)
    usleep(100);
  changeend = 1;
  switch (nextlooptrig) {
    case 0:
      if (oldonce) {
	disableLooping();
	oldonce = 0;
	strig = 0;
	while (changestate)
	  usleep(100);
	oldstate = -1;
	nexti = currentloopindex;
	nexts = currentloopsession;
	newstate = 0;
	changestate = 1;
      }
      process_sl_trigger();
      break;
    case 1:
      if (!loops[nexti][nexts].start)
	firststartbridge(&loops[nexti][nexts]);
      switch (nextsm) {
	case 0:
	  oldonce = 0;
	  strig = 0;
	  nextlooptrig = 0;
	  while (changestate)
	    usleep(100);
	  oldstate = 0;
	  changestate = 1;
	  break;
	case 1:
	  oldonce = 0;
	  strig = 1;
	  break;
	case 2:
	  oldonce = 1;
	  strig = 0;
	  nextlooptrig = 2;
	  while (changestate)
	    usleep(100);
	  oldstate = 0;
	  newstate = 1;
	  changestate = 1;
	  break;
      }
      switch (nextslm) {
	case 0:
	  slrec = 0;
	  sloverd = 0;
	  slmulti = 0;
	  newstate = 1;
	  break;
	case 1:
	  slrec = 1;
	  sloverd = 0;
	  slmulti = 1;
	  newstate = 2;
	  break;
	case 2:
	  slrec = 1;
	  sloverd = 0;
	  slmulti = 0;
	  newstate = 1;
	  break;
	case 3:
	  slrec = 1;
	  sloverd = 1;
	  slmulti = 0;
	  newstate = 1;
	  break;
	case 4:
	  slrec = 1;
	  sloverd = 1;
	  slmulti = 1;
	  newstate = 2;
	  break;
      }
      if (nextsm == 2) {
	  slmulti = 0;
	  newstate = 1;
      }
      setCurrentLoop(loops[nexti][nexts]);
      break;
    case 2:
      disableLooping();
      oldonce = 0;
      strig = 0;
      while (changestate)
	usleep(100);
      oldstate = -1;
      nexti = currentloopindex;
      nexts = currentloopsession;
      newstate = 0;
      changestate = 1;
      process_sl_trigger();
      break;
  }
  if (nextlooptrig == 1) {
    while (changestate)
      usleep(100);
    nextlooptrig = 0;
    oldstate = 0;
    changestate = 1;
  }
  changeend = 0;
}

int endasync() {
  while (changeend)
    usleep(100);
  changeend = 1;
  int res = 0;
  if (asyncEnabled > 0 && loops[asyncLoopindex][asyncLoopsession].end == 1) {
    res = 1;
    asyncEnabled = 0;
    resetbridge(&loops[asyncLoopindex][asyncLoopsession]);
    int ni, ns, nni, nns, os;
    ni = currentloopindex;
    ns = currentloopsession;
    nni = nexti;
    nns = nexts;
    os = oldstate;
    while (changestate)
      usleep(100);  
    nexti = asyncLoopindex;
    nexts = asyncLoopsession;
    oldstate = -1;
    newstate = 0;  
    changestate = 1;
    while (changestate)
      usleep(100);  
    nexti = nni;
    nexts = nns;
    currentloopindex = ni;
    currentloopsession = ns;
    currentindex = loops[ni][ns].index;
    currentsession = loops[ni][ns].session;
    oldstate = os;  
  }
  changeend = 0;
  return res;
}

void process_sl_trigger() {
  if (debug) {
    fprintf(stdout, "Process SL trigger: slrec %i, sloverd %i, slmulti %i \n\t\ttslrec %i, tsloverd %i, tslmulti %i\n", slrec, sloverd, slmulti, tslrec, tsloverd, tslmulti);
    //fprintf(stdout, "\t\tnrec %i, nmulti %i\n", nrec, nmulti);
  }
  int cmd = 4;
  
  if (((slrec == tslrec) && (slrec)) || (slrec && tsloverd)) {
    slrec = 0;
    tslrec = 0;
    tslmulti = 0;
    if (tsloverd) {
      cmd = 1;
      tsloverd = 0;
    } else {
      cmd = 0;
    }
  }
  
  if (slrec || tslrec) {
    if (tslrec) {
      tslrec = 0;
      if (tsloverd) {
	if (slmulti) {
	  tslmulti++;
	  tslrec = 1;
	} else {
	  cmd = 1;
	  tsloverd = 2;
	}
	sloverd = 0;
      } else {
	if (slmulti) {
	  tslmulti++;
	  tslrec = 1;
	} else {
	  tslmulti = 0;
	  cmd = 0;
	}
      }
    } else {
      if (sloverd) {
	tsloverd = 1;
      } else {
	
      }
      slrec = 0;
      cmd = 0;
      tslrec = 1;
    }      
  }

  if ((tsloverd == 2) && (cmd == 4)) {
    if (!tslmulti) {
      tsloverd = 0;
      cmd = 1;
    } else {
      tslmulti--;
    }
  }
  
  if (cmd < 4)
    send_midi_sl(cmd, 0);
}

void send_midi_sl(int type, int val) {
  /* Standard MIDI channel for sooperlooper */
  snd_seq_event_t ev;
  snd_seq_ev_clear(&ev);
  snd_seq_ev_set_direct(&ev);
  snd_seq_ev_set_source(&ev, midi_out);
  snd_seq_ev_set_subs(&ev);
  switch (type) {
    case 0: /* Record */
      snd_seq_ev_set_noteon(&ev, 15, 0, 127);
      break;      
    case 1: /* Multi */
      snd_seq_ev_set_noteon(&ev, 15, 1, 127);
      break;	      
    case 2: /* Pause */
      snd_seq_ev_set_noteon(&ev, 15, 2, 127);
      break;	  
    case 3: /* Volume */
      snd_seq_ev_set_controller(&ev, 15, 0, val);
      break;	  
  }
  snd_seq_event_output_direct(seq, &ev);
  snd_seq_drain_output(seq);
}