/* cyclonic.c
 *
 *
 *    this file is part of:
 *
 *    Cyclone   ---   Loop Giant
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Sergio Atzori, 2010
 *
 *
 */

#include <math.h>

#include "slconnect.h"
#include "slverify.h"
#include "cyclonic.h"
#include "bridge.h"

/*
 * Functions to retrieve/store states
 * needed to break infinite cycle
 */

char *cyclonic_msg() {
  return (char *)&cymsg;
}

void cyclonic_stop() {
  /* set bridge.c */
  currentEnabled = 0;
  asyncEnabled = 0;  
  
  if (seqcy)
    snd_seq_close(seqcy);
  if (jack_deactivate(client) == 0)
    jack_client_close(client);
  cystate = -1;
  verify_trigger = 0;
  strcpy(cymsg, "Jack1");
  if (cydebug)
   fprintf (stdout, "Jack Client interrupted by user\n");
}

void cyclonic_kill_midi() {
  if (seqcy)
    snd_seq_close(seqcy);  
}
  
void jack_shutdown(void *arg)
{
  /* set bridge.c */
  currentEnabled = 0;
  asyncEnabled = 0;  
  
  if (jack_deactivate(client) == 0)
    jack_client_close(client);
  cystate = 1;
  verify_trigger = 0;
  strcpy(cymsg, "Jack0");
  fprintf (stderr, "Jack kicked all of us out!\n");
}

int cyclonic_state() {
  return cystate;
}

long cyclonic_samplerate(jack_client_t *client) {
  return (long) jack_get_sample_rate(client);
}

int async_process(jack_nframes_t nframes) {
  if (asyncEnabled == 0) {
    /* Silence */
    int i;
    for(i = 0; i < (long)nframes; i++) {
      async_out1[i] = 0.0;
      async_out2[i] = 0.0;
    }
  } else {
    int nsamples = nframes * asyncLoop.channels;
    int nbytes = nsamples * sizeof(float);
    
    
    /* Ensure the period size is workable. */
    if(nbytes >= asyncLoop.buffer_bytes) {
      strcpy(cymsg, "Jack5");
      fprintf(stderr, "CYCLONIC: period size exceeds limit\n");
      return 1;
    }
    
    /* Get data from sample rate converter, this returns the number of frames acquired. */
    double ratio = (double)jack_get_sample_rate(client)/(double)asyncLoop.sfnfo.samplerate;
    long answer = src_callback_read(asyncLoop.src, ratio, (long)nframes, asyncLoop.j_buffer);
    if(answer == 0) {
      fprintf(stderr, "CYCLONIC: sample rate converter failed: %s\n", src_strerror(src_error(asyncLoop.src)));
      return 1;
    }
    
    int _i;
    float volume = quickavolume/(float)100;
    
    /* Uninterleave available data to the output buffers. */    
    for(_i = 0; _i < answer; _i++) {
      /* if mono loop, duplicate signal, else use channels 1,2 from file */
      if (asyncLoop.channels == 1) { /* mono */
	async_out1[_i] = asyncLoop.j_buffer[(_i*asyncLoop.channels)] * volume;
	async_out2[_i] = asyncLoop.j_buffer[(_i*asyncLoop.channels)] * volume;	
      } else {
	async_out1[_i] = asyncLoop.j_buffer[(_i*asyncLoop.channels)] * volume;
	async_out2[_i] = asyncLoop.j_buffer[(_i*asyncLoop.channels) + 1] * volume;
      }
    }
    
    /* If any sample data is unavailable inform the user and zero the
      output buffers.  The print statement is not correct, a this
      should set a flag and have another thread take appropriate
      action. */
    if(answer < (int)nframes) {
      fprintf(stderr, "CYCLONIC: disk thread late (%ld < %d)\n", answer, nframes);
      for(_i = answer; _i < (int)nframes; _i++) {
	async_out1[_i] = 0.0;
	async_out2[_i] = 0.0;	
      }
    }

    /* Indicate to the disk thread that the ring buffer has been read
      from.  This is done by writing a single byte to a communication
      pipe.  Once the disk thread gets so far ahead that the ring
      buffer is full it reads this communication channel to wait for
      space to become available.  So long as PIPE_BUF is not a
      pathologically small value this write operation is atomic and
      will not block.  The number of bytes that can accumulate in the
      pipe is a factor of the relative sizes of the ring buffer and the
      process callback, but should in no case be very large. */

    char b = 1;
    int writepipe = write(asyncLoop.pipe[1], &b, 1);
    if(writepipe == -1) {
      fprintf(stderr, "CYCLONIC: write() into pipe failed\n");
      return 1; 
    }
  }
  return 0;      
}


/*
 * The process callback for this JACK application.
 * It is called by JACK at the appropriate times.
 */
int process_buffer(jack_nframes_t nframes, void *arg)
{
  int asp = async_process(nframes);
  if (asp) {
    if (strcmp(cymsg,"Jack5") != 0) {
      strcpy(cymsg, "Jack6");
      fprintf(stderr, "CYCLONIC: trouble within async process\n");
    }
    return 1;
  }
    
  struct loopdata *loop = (struct loopdata *)arg;
  jack_default_audio_sample_t *out1 = (jack_default_audio_sample_t *) jack_port_get_buffer (output_port1, nframes);
  jack_default_audio_sample_t *out2 = (jack_default_audio_sample_t *) jack_port_get_buffer (output_port2, nframes);
  
  if (currentEnabled == 0) {
    /* Silence */
    int i;
    float lmax = 0;
    float rmax = 0;
    float genvolume = (float)gvolume/(float)100;
    
    for(i = 0; i < (long)nframes; i++) {
      out1[i] = async_out1[i] * genvolume;
      out2[i] = async_out2[i] * genvolume;
      if (fabs((float)out1[i]) > lmax) {
	lmax = fabs((float)out1[i]);
      }
      if (fabs((float)out2[i]) > rmax) {
	rmax = fabs((float)out2[i]);
      }
    }
    
    lsig = (int)(pow(lmax,1.0/4.0) * (float)256);
    rsig = (int)(pow(rmax,1.0/4.0) * (float)256);
  } else {
    int nsamples = nframes * loop->channels;
    int nbytes = nsamples * sizeof(float);
    
    
    /* Ensure the period size is workable. */
    if(nbytes >= loop->buffer_bytes) {
      strcpy(cymsg, "Jack5");
      fprintf(stderr, "CYCLONIC: period size exceeds limit\n");
      return 1;
    }
    
    /* Get data from sample rate converter, this returns the number of frames acquired. */
    double ratio = (double)jack_get_sample_rate(client)/(double)loop->sfnfo.samplerate;
    long answer = src_callback_read(loop->src, ratio, (long)nframes, loop->j_buffer);
    if(answer == 0) {
      fprintf(stderr, "CYCLONIC: sample rate converter failed: %s\n", src_strerror(src_error(loop->src)));
      return 1;
    }
        
    /* Uninterleave available data to the output buffers. */
    int _i;
    float loopvolume = quick_volume(loop)/(float)100;
    float genvolume = (float)gvolume/(float)100;
    
    float lmax = 0;
    float rmax = 0;
    
    for(_i = 0; _i < answer; _i++) {
      /* if mono loop, duplicate signal, else use channels 1,2 from file */
      if (loop->channels == 1) { /* mono */
	out1[_i] = ((loop->j_buffer[(_i*loop->channels)] * loopvolume) + async_out1[_i]) * genvolume;
	out2[_i] = ((loop->j_buffer[(_i*loop->channels)] * loopvolume) + async_out2[_i]) * genvolume;	
      } else {
	out1[_i] = ((loop->j_buffer[(_i*loop->channels)] * loopvolume) + async_out1[_i]) * genvolume;
	out2[_i] = ((loop->j_buffer[(_i*loop->channels) + 1] * loopvolume) + async_out2[_i]) * genvolume;
      }
      if (fabs((float)out1[_i]) > lmax) {
	lmax = fabs((float)out1[_i]);
      }
      if (fabs((float)out2[_i]) > rmax) {
	rmax = fabs((float)out2[_i]);
      }
    }
    
    lsig = (int)(pow(lmax,1.0/4.0) * (float)256);
    rsig = (int)(pow(rmax,1.0/4.0) * (float)256);

    /* If any sample data is unavailable inform the user and zero the
      output buffers.  The print statement is not correct, a this
      should set a flag and have another thread take appropriate
      action. */
    if(answer < (int)nframes) {
      fprintf(stderr, "CYCLONIC: disk thread late (%ld < %d)\n", answer, nframes);
      for(_i = answer; _i < (int)nframes; _i++) {
	out1[_i] = 0.0;
	out2[_i] = 0.0;	
      }
    }

    /* Indicate to the disk thread that the ring buffer has been read
      from.  This is done by writing a single byte to a communication
      pipe.  Once the disk thread gets so far ahead that the ring
      buffer is full it reads this communication channel to wait for
      space to become available.  So long as PIPE_BUF is not a
      pathologically small value this write operation is atomic and
      will not block.  The number of bytes that can accumulate in the
      pipe is a factor of the relative sizes of the ring buffer and the
      process callback, but should in no case be very large. */

    char b = 1;
    int writepipe = write(loop->pipe[1], &b, 1);
    if(writepipe == -1) {
      fprintf(stderr, "CYCLONIC: write() into pipe failed\n");
      return 1; 
    }
  }
  return 0;      
}

int cyclonic(char *nomeclient) {
  /* set bridge.c */
  currentEnabled = 0;
  asyncEnabled = 0;
  oldstate = -1;

  sooloovol =  0;
  
  setDebugBridge(debug);
  char cymidi[6];
  const char **ports;
  jack_options_t opz = JackNoStartServer;
  int midi_out_id;

  if (strlen(nomeclient) < 1) {
	  cystate = 1;
	  strcpy(cymsg, "General0");
	  if (cydebug)
	    fprintf (stderr, "Error: no name supplied for Jack/Alsa clients\n");
	  return 1;
  }

  /* try to become a client of the JACK server and ALSA seq */

  if ((client = jack_client_open (nomeclient, opz, NULL)) == 0) {
	  cystate = 1;
	  strcpy(cymsg, "Jack2");
	  if (cydebug)	
	    fprintf (stderr, "Jack server not running?\n");
	  return 1;
  }

  if (snd_seq_open(&seqcy, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0) {
      cystate = 1;
      strcpy(cymsg, "Alsa0");
      if (cydebug)
	fprintf (stderr, "Error opening ALSA sequencer.\n");
      return 1;
  }

  /* tell the JACK server to call `process()' whenever
      there is work to be done.
  */
  asyncLoop.index = 128;
  asyncLoop.session = 128;
  jack_set_process_callback (client, process_buffer, &currentLoop);

  /* tell the JACK server to call `jack_shutdown()' if
      it ever shuts down, either entirely, or if it
      just decides to stop calling us.
  */

  jack_on_shutdown (client, jack_shutdown, 0);


  snd_seq_set_client_name(seqcy, nomeclient);
	  
  if ((midi_out_id = snd_seq_create_simple_port(seqcy, nomeclient,
					  SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ,
					  SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
		  cystate = 1;
		  strcpy(cymsg, "Alsa1");
		  if (cydebug)
		    fprintf (stderr, "Error creating sequencer input port.\n");
		  return 1;
  }

  sprintf(cymidi,"%i:0",snd_seq_client_id(seqcy));

  if ((midi_in_id = snd_seq_create_simple_port(seqcy, nomeclient,
					  SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE,
					  SND_SEQ_PORT_TYPE_APPLICATION)) < 0) {
		  cystate = 1;
		  strcpy(cymsg, "Alsa2"); 
		  if (cydebug)
		    fprintf (stderr, "Error creating sequencer output port.\n");
		  return 1;
  }

  /*
  input_port1 = jack_port_register (client, "in1", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
  input_port2 = jack_port_register (client, "in2", JACK_DEFAULT_AUDIO_TYPE, JackPortIsInput, 0);
  */
  output_port1 = jack_port_register (client, "out1", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);
  output_port2 = jack_port_register (client, "out2", JACK_DEFAULT_AUDIO_TYPE, JackPortIsOutput, 0);

  /* tell the JACK server that we are ready to roll */

  if (jack_activate (client)) {
	  cystate = 1;
	  strcpy(cymsg, "Jack3");
	  if (cydebug)
	    fprintf (stderr, "Cannot activate Jack client\n");
	  return 1;
  }

  /* connect the ports. Note: you can't do this before
      the client is activated, because we can't allow
      connections to be made to clients that aren't
      running.
  */

  /*
  if ((ports = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsOutput)) == NULL) {
	  cystate = 1;
	  strcpy(cymsg, "");
	  strcat(cymsg, "Cannot find any physical capture ports\n");
	  fprintf (stderr, cymsg);
	  strcat(cymsgs,cymsg);
	  return 1;
  }

  if (jack_connect (client, ports[0], jack_port_name (input_port1))) {
	  cystate = 1;
	  strcat(cymsgs,cymsg);
	  strcat(cymsg, "Cannot connect input port 1\n");
	  fprintf (stderr, cymsg);
	  strcpy(cymsg, "");
  }
  if (jack_connect (client, ports[1], jack_port_name (input_port2))) {
	  cystate = 1;
	  strcpy(cymsg, "");
	  strcat(cymsg, "Cannot connect input port 2\n");
	  fprintf (stderr, cymsg);
	  strcat(cymsgs,cymsg);
  }

  free (ports);
  */

  if ((ports = jack_get_ports (client, NULL, NULL, JackPortIsPhysical|JackPortIsInput)) == NULL) {
	  cystate = 1;
	  strcpy(cymsg, "");
	  strcat(cymsg, "Jack4");
	  if (cydebug)
	    fprintf (stderr, "Cannot find any physical playback ports\n");
	  return 1;
  }

  if (jack_connect (client, jack_port_name (output_port1), ports[0])) {
	  fprintf (stderr, "Cannot connect output port 1\n");
  }

  if (jack_connect (client, jack_port_name (output_port2), ports[1])) {
	  fprintf (stderr, "Cannot connect output port 2\n");
  }

  free (ports);

  cystate = 0;
  /* Attempt to connect Cyclone MIDI to sooperlooper 
  sl_connect();*/


  if (debug)
    fprintf (stdout, "Jack Client started\n");

  /* set bridge.c */
  setSamplerate(cyclonic_samplerate(client));
  setJackClient(client);
  setMidiSeq(seqcy, midi_out_id);
  
  return 0;
}

int sl_connect() {
  if (cystate == 1)
    return 0;
  sprintf(cymidi,"%i:0",snd_seq_client_id(seqcy));
  strncpy(slmidi,slconnect(cymidi),6);
  if ( strcmp(slmidi,"0") == 0 ) {
    strcpy(cymsg, "Alsa3");
    if (cydebug)
      fprintf(stderr, "Cannot connect to Sooperlooper\n");
    verify_trigger = 0;
    return 1;
  } else {
    strcpy(cymsg, "Alsa4");
    if (cydebug)
      fprintf(stdout, "Connected to Sooperlooper\n");
    verify_trigger = 1;
  }
  return 0;
}

int sl_verify() {
  if (cystate == 1)
    return 0;
  sprintf(cymidi,"%i:0",snd_seq_client_id(seqcy));
  if (verify_trigger == 1) {
    if (slverify(seqcy,cymidi,slmidi) != 0) {
      strcpy(cymsg, "Alsa5");
      if (cydebug)
	fprintf(stdout, "Disconnected from Sooperlooper\n");
      verify_trigger = 0;
      return 1;
    }
  }
  return 0;
}

void cyclonic_stop_looping() {
  totallyDisableLooping();
  send_midi_sl(2, 0);
  send_midi_sl(4, sooloovol);
}

void setBufferFramesCyclonic(int bf) {
  buffer_frames = bf;  
}

void setQuickFadeCyclonic(int qf) {
  quickmsec = qf;  
}

void setDebugCyclonic(int dbg) {
  debug = dbg;  
}

void set_general_volume(int gvol) {
  gvolume = gvol;
}

void set_standard_coord(int ind, int ses) {
  currentEnabled = 0;
  asyncEnabled = 0;  
  oldstate = -1;
  currentindex = ind;
  currentsession = ses;
  oldonce = 0;
  strig = 0;
  nextlooptrig = 0;
  changestate = 0;
  nexti = 128;
  nexts = 128;
  slrec = 0;
  sloverd = 0;
  slmulti = 0;
  tslrec = 0;
  tsloverd = 0;
  tslmulti = 0;
}

struct statechange check_bridge() {
  struct statechange sc;
  sc.changed = changestate;
  if (changestate) {
    changestate = 0;
    sc.oldi = loops[oldindex][oldsession].index;
    sc.olds = loops[oldindex][oldsession].session;
    sc.oldstate = oldstate;
    currentloopindex = nexti;
    currentloopsession = nexts;
    currentindex = loops[nexti][nexts].index;
    currentsession = loops[nexti][nexts].session;
    sc.newi = currentindex;
    sc.news = currentsession;
    sc.newstate = newstate;
  }
  
  return sc;
}

int position() {
  int ps;
  if (currentEnabled == 1) {
    ps = (int)((((float)currentposition) / (float)currentsamples) * (float)256);
  } else {
    ps = 0;
  }
  return ps;
}

int async() {
  return asyncEnabled;
}

int asyncposition() {
  int ps;
  if (asyncEnabled == 1) {
    ps = (int)((((float)currentasyncposition) / (float)loops[asyncLoopindex][asyncLoopsession].sfnfo.frames) * (float)256);
  } else {
    ps = 0;
  }
  return ps;
}

int signalL() {
  int sl = 0;
  if (currentEnabled || asyncEnabled) {
    sl = lsig - 5;
  }
  
  return sl;
}

int signalR() {
  int sr = 0;
  if (currentEnabled || asyncEnabled) {
    sr = rsig - 5;
  }
  
  return sr;
}

void setsooloovol(int vol) {
  sooloovol = vol;
  if (cystate == 1)
    return;
  send_midi_sl(3, vol);
}