/* slconnect.c
 *
 *
 *    this file is part of:
 *
 *    Cyclone   ---   Loop Giant
 *
 * MIDI function library to connect
 * calling application to sooperlooper
 *
 * taken from
 * >aconnect.c
 * >
 * >connect / disconnect two subscriber ports
 * >  ver.0.1.3
 * >
 * >Copyright (C) 1999 Takashi Iwai
 * >
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 *  -----------------------------------------------------------
 *
 *  Freely modified by Sergio Atzori, 2010
 *
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/ioctl.h>
#include <alsa/asoundlib.h>


static void error_handler(const char *file, int line, const char *function, int err, const char *fmt, ...)
{
	va_list arg;

	if (err == ENOENT)	/* Ignore those misleading "warnings" */
		return;
	va_start(arg, fmt);
	fprintf(stderr, "ALSA lib %s:%i:(%s) ", file, line, function);
	vfprintf(stderr, fmt, arg);
	if (err)
		fprintf(stderr, ": %s", snd_strerror(err));
	putc('\n', stderr);
	va_end(arg);
}

/*
 * check permission (capability) of specified port
 */

#define LIST_INPUT	1
#define LIST_OUTPUT	2

#define perm_ok(pinfo,bits) ((snd_seq_port_info_get_capability(pinfo) & (bits)) == (bits))

static int check_permission(snd_seq_port_info_t *pinfo, int perm)
{
	if (perm) {
		if (perm & LIST_INPUT) {
			if (perm_ok(pinfo, SND_SEQ_PORT_CAP_READ|SND_SEQ_PORT_CAP_SUBS_READ))
				goto __ok;
		}
		if (perm & LIST_OUTPUT) {
			if (perm_ok(pinfo, SND_SEQ_PORT_CAP_WRITE|SND_SEQ_PORT_CAP_SUBS_WRITE))
				goto __ok;
		}
		return 0;
	}
 __ok:
	if (snd_seq_port_info_get_capability(pinfo) & SND_SEQ_PORT_CAP_NO_EXPORT)
		return 0;
	return 1;
}

/*
 * search all ports
 */

static int do_search_port(snd_seq_t *seq, int perm)
{
	snd_seq_client_info_t *cinfo;
	snd_seq_port_info_t *pinfo;
	int count;

	snd_seq_client_info_alloca(&cinfo);
	snd_seq_port_info_alloca(&pinfo);
	snd_seq_client_info_set_client(cinfo, -1);
	while (snd_seq_query_next_client(seq, cinfo) >= 0) {
		/* reset query info */
		snd_seq_port_info_set_client(pinfo, snd_seq_client_info_get_client(cinfo));
		snd_seq_port_info_set_port(pinfo, -1);
		count = 0;
		while (snd_seq_query_next_port(seq, pinfo) >= 0) {
			if (check_permission(pinfo, perm)) {
				if (strcmp((snd_seq_client_info_get_name(cinfo)), "sooperlooper") == 0) return snd_seq_client_info_get_client(cinfo);
				count++;
			}
		}
	}
	return -1;
}

char *slconnect(char *port)
{
	snd_seq_t *seq;
	int queue = 0, convert_time = 0, convert_real = 0, exclusive = 0;
	int list_perm = 0;
	int client;
	int slport;
	char slportc[6];
	snd_seq_port_subscribe_t *subs;
	snd_seq_addr_t sender, dest;
	list_perm |= LIST_OUTPUT;
	
	if (snd_seq_open(&seq, "default", SND_SEQ_OPEN_DUPLEX, 0) < 0) {
		fprintf(stderr, "can't open sequencer\n");
		return "0";
	}

	snd_lib_error_set_handler(error_handler);

	slport = do_search_port(seq, list_perm);
	
	if (slport < 0) {
	  fprintf(stderr, "SooperLooper not found\n");
	  return "0";
	}
	
	sprintf(slportc,"%i:0",slport);
	
	if ((client = snd_seq_client_id(seq)) < 0) {
		snd_seq_close(seq);
		fprintf(stderr, "can't get client id\n");
		return "0";
	}

	/* set client info */
	if (snd_seq_set_client_name(seq, "ALSA Connector") < 0) {
		snd_seq_close(seq);
		fprintf(stderr, "can't set client info\n");
		return "0";
	}

	/* set subscription */
	if (snd_seq_parse_address(seq, &sender, port) < 0) {
		snd_seq_close(seq);
		fprintf(stderr, "invalid sender address %s\n", port);
		return "0";
	}
	
	if (snd_seq_parse_address(seq, &dest, slportc) < 0) {
		snd_seq_close(seq);
		fprintf(stderr, "invalid destination address %s\n", slportc);
		return "0";
	}
	
	snd_seq_port_subscribe_alloca(&subs);
	snd_seq_port_subscribe_set_sender(subs, &sender);
	snd_seq_port_subscribe_set_dest(subs, &dest);
	snd_seq_port_subscribe_set_queue(subs, queue);
	snd_seq_port_subscribe_set_exclusive(subs, exclusive);
	snd_seq_port_subscribe_set_time_update(subs, convert_time);
	snd_seq_port_subscribe_set_time_real(subs, convert_real);

	if (snd_seq_get_port_subscription(seq, subs) == 0) {
		snd_seq_close(seq);
		fprintf(stderr, "Sooperlooper is already connected\n");
		return slportc;
	}
	if (snd_seq_subscribe_port(seq, subs) < 0) {
		snd_seq_close(seq);
		fprintf(stderr, "Connection failed (%s)\n", snd_strerror(errno));
		return "0";
	}

	snd_seq_close(seq);

	return slportc;
}
